"""
The Python CA engine.

Xentica is the framework to build GPU-accelerated models for
multi-dimensional cellular automata. Given pure Python definitions, it
generates kernels in CUDA C and runs them on NVIDIA hardware.

"""
